from deco import model
import pysnooper


@pysnooper.snoop("./debug.log", depth=1000)
def main():
    class Decorator(model.Deco_Expand):
        def begin(self, func, args, kwargs):
            if func(*args, **kwargs) != (args, kwargs):
                raise TypeError("Cannot modify params")

    @Decorator()
    def modify(*args, **kwargs):
        return kwargs, args

    x, y = modify(1, 2)
    print(x, y)


if __name__ == "__main__":
    main()



# coding = utf_8
# NiceDecorator.decorator
"""
module NiceDecorator.decorator.

A decoration module(Decorator)

class:
    Base_Deco(object): Decorator's base class
    group Deco_for_function:
        Deco_Function(Base_Deco): Decorated function Decorator's base class
        Deco_aFunction(Deco_Function): A decorator template class (pass in method)
        Deco_iFunction(Deco_Function): A decorator template class (inheritance)
        Deco_wFunction(Base_Deco): Write Wrapper Return a Decorator
    group Deco_for_class:
        Deco_Class(Base_Deco): Decorated class decorator's object class
        Deco_aClass(Deco_Class): Decorated a Class decorator (pass in method)
        Deco_iClass(Deco_Class): Decorated a Class decorator (inheritance)
    group Deco_combination:
        Deco_Union(Base_Deco): Union the functions
        Deco_List(Base_Deco): List the functions

"""
import functools
from typing import Any, Callable
from . import PASS, set_param, FUNCTION, EMPTY


class Base_Deco(object):
    """
    class NiceDecorator.decorator.Base_Deco(object)

    Decorator's base class

    No use directly!
    """
    def __str__(self):
        return f"Decorator object: {self.__class__}"


# {
class Deco_Function(Base_Deco):
    """
    class NiceDecorator.decorator.Deco_Function(Base_Deco)

    Decorated function decorator's object class

    Please do not use directly!
    """
    
    def __call__(self, func: Callable, /, **c):
        def decofunc(func):
            """
            function decofunc(func)

            A decorator function

            params:
                func: Decorated function

            return:
                wrapper: Decorated function
            """
            nonlocal c
            self._param_it = self._param
            self._param_it.update(c)

            # Action on definition
            _d, wa = set_param.setparam(self._d, self._param_it, "d")
            if "func" in wa:
                func = _d(func=func)
            else:
                _d()

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                _b, wa1 = set_param.setparam(self._b, self._param_it, "b")
                _e, wa2 = set_param.setparam(self._e, self._param_it, "e")

                # Operation before operation
                if "args" in wa1:
                    if "kwargs" in wa1:
                        if "tram" in wa2:
                            args, kwargs, tram = _b(args=args, kwargs=kwargs)
                        else:
                            args, kwargs = _b(args=args, kwargs=kwargs)
                    else:
                        if "tram" in wa2:
                            args, tram = _b(args=args)
                        else:
                            args = _b(args=args)
                else:
                    if "kwargs" in wa1:
                        if "tram" in wa2:
                            kwargs, tram = _b(kwargs=kwargs)
                        else:
                            kwargs = _b(kwargs=kwargs)
                    else:
                        if "tram" in wa2:
                            tram = _b()
                        else:
                            _b()

                res = func(*args, **kwargs)  # Run

                # Operation after operation
                if "tram" in wa2:
                    if "res" in wa2:
                        res = _e(tram=tram, res=res)
                    else:
                        _e(tram=tram)
                else:
                    if "res" in wa2:
                        res = _e(res=res)
                    else:
                        _e()
                return res

            return wrapper
        
        if self.pck:
            return decofunc
        else:
            return decofunc(func)

    @property
    def param(self):
        return self._param

    @property
    def pck(self):
        return self._pck


class Deco_aFunction(Deco_Function):
    """
    class NiceDecorator.decorator.Deco_aFunction(Deco_Function)

    A decorator template class (pass in method)

    method:
        __init__:
            params:
                adef: do some before definition [must be a function] (NiceDecorator.PASS)
                wrap: do some in run [must be a tuple[function, NiceDecorator.FUNCTION, function]] ( (NiceDecorator.PASS, NiceDecorator.FUNCTION, NiceDecorator.PASS) )
                pck: is pack (False)
                --other-- -> **kwargs: set some params(global)

        __call__:
            params:
                --other-- -> **c: set some params(part)

            return:
                function decofunc(func):
                    params:
                        func: Decorated function

                    return:
                        wrapper: Decorated function

                (if not pck, call = decofunc)

    self params:
        wrap: wrapper
        param: global params
        pck: is packed

    __str__ -> 'Decorator object: {self.__class__}'
    
    for example:
    >>> decor = decorator.Deco_aFunction(wrap=(timecount.time_start_ns, FUNCTION, timecount.time_end_ns))
    >>> @decor
    >>> def f():
    ...     print(1)
    ...
    >>> a = f()
    >>> b = f()
    >>> print(a, b)
    181000 168600
    
    """

    def __init__(self, adef=PASS, wrap=(PASS, FUNCTION, PASS), pck=False, **kwargs):
        assert wrap[1] == FUNCTION
        self._d = adef
        self._b = wrap[0]
        self._e = wrap[2]
        self._param = kwargs
        self._pck = pck
        self._param_it = self._param


class Deco_iFunction(Deco_Function):
    """
    class NiceDecorator.decorator.Deco_iFunction(Deco_Function)

    A decorator template class (inheritance)

    method:
        __init__:
            params:
                pck: is pack (False)
                --other-- -> **kwargs: set some params(global)

        __call__:
            params:
                --other-- -> **c: set some params(part)

            return:
                function decofunc(func):
                    params:
                        func: Decorated function

                    return:
                        wrapper: Decorated function

                (if not pck, call = decofunc)

        (definition adef,begin,end)

    self params:
        wrap: (self.begin, NiceDecorator.FUNCTION, self.end)
        param: global params
        pck: is packed

    __str__ -> Decorator object: {self.__class__}
    
    for example:
    >>> class DECORATOR(decorator.Deco_iFunction):
    ...     def begin(self):
    ...          print("start")
    ...     def end(self):
    ...        print("end")
    ...     def adef(self):
    ...         print("def")
    ...
    >>> @DECORATOR()
    >>> def g():
    ...     print(2)
    ...
    def
    >>> g()
    start
    2
    end
    >>> g()
    start
    2
    end
    
    """

    def __init__(self, pck=False, **kwargs):
        self._d = self.adef
        self._b = self.begin
        self._e = self.end
        self._param = kwargs
        self._param_it = self._param
        self._pck = pck

    def begin(self):
        """
        method begin(self) (in NiceDecorator.decorator.Deco_iFunction)

        pass
        """
        pass

    def end(self):
        """
        method end(self) (in NiceDecorator.decorator.Deco_iFunction)

        pass
        """
        pass

    def adef(self):
        """
        method adef(self) (in NiceDecorator.decorator.Deco_iFunction)

        pass
        """
        pass

    @property
    def wrap(self):
        return (self.adef, self.begin, self.end)


class Deco_wFunction(Base_Deco):
    """
    class Deco_wFunction(Base_Deco)

    Write Wrapper Return a Decorator

    method:
        __init__:
            params:
                func: Decorated function

        __call__:
            params:
                if with paramters:
                    --other-- -> *args, **kwargs: paramters in decorator
                else:
                    f: Decorated function

            return:
                res: Decorated function's response

        __getattr__:
            return:
                self.func.{attribute}

    self params:
        func: decorator
        param_type: is with params ('With' or 'Without')

    __str__ -> 'Decorator object: {self.__class__}'

    for example:
    >>> @decorator.Deco_wFunction
    >>> def decor2(func, *args, **kwargs):
    ...     print("===")
    ...     res = func(*args, **kwargs)
    ...     print("===")
    ...     return res
    ...
    >>> @decor2
    >>> def h():
    ...     print(3)
    ...
    >>> h()
    ===
    3
    ===
    >>> h()
    ===
    3
    ===
    
    """

    def __init__(self, func: Callable):
        arg = set_param.getparam(func)
        param = []
        for _ in arg[1:]:
            if _[0] != "*":
                param.append(_)
        if len(param) > 0:
            @functools.wraps(func)
            def bz(**kw):
                def decorator(f):
                    cc_d = {}
                    for _1, _2 in kw.items():
                        if _1 in param:
                            cc_d[_1] = _2

                    @functools.wraps(f)
                    def wrapper(*listt, **dictt):
                        res = func(f, *listt, **dictt, **cc_d)
                        return res

                    return wrapper

                return decorator

            self._func = bz
            self._type = True
        else:
            @functools.wraps(func)
            def decor(f):
                @functools.wraps(f)
                def wrapper(*listt, **dictt):
                    res = func(f, *listt, **dictt)
                    return res

                return wrapper

            self._func = decor
            self._type = False
        
    def __call__(self, f: Callable, *args: Any, **kwargs):
        if self._type:
            res = self._func(*args, **kwargs)
            return res
        else:
            res = self._func(f)
            return res

    @property
    def func(self):
        return self._func

    @property
    def param_type(self):
        if self._type:
            return "With"
        else:
            return "Without"

    def __getattr__(self, item):
        return getattr(self._func, item)

# }Deco_for_function


# {
class Deco_Class(Base_Deco):
    """
    class NiceDecorator.decorator.Deco_Class(Base_Deco)

    Decorated class decorator's object class

    Please do not use directly!
    """
    def __call__(self, cls=EMPTY, *m, **c):
        def decofunc(cls):
            """
            function decofunc(cls)

            A decorator function

            params:
                cls: Decorated class

            return:
                cls: New class
            """
            self._param_it = self._param
            self._method_it = self._method
            
            # Append methods
            self._method_it.extern(m)
            for f in self._method:
                f_name = f.__name__
                if f_name[0:2] == "__" and f_name[-2:] != "__":
                    f_name = "_" + cls.__name__ + "_" + f_name[3:]  # Change method name during compilation
                setattr(cls, f_name, f)
                
            # Appand attributes
            self._param_it.update(c)
            for p, p_name in self._param_it.items():
                if p_name[0:2] == "__" and p_name[-2:] != "__":
                    p_name = "_" + cls.__name__ + "_" + p_name[3:]  # Change attribute name during compilation
                setattr(cls, p_name, p)
            return cls

        if self._pck:
            return decofunc
        else:
            return decofunc(cls)

    @property
    def pck(self):
        return self._pck

    @property
    def param(self):
        return self._param
    
    @property
    def method(self):
        return self._method


class Deco_aClass(Base_Deco):
    """
    class NiceDecorator.decorator.Deco_aClass(NiceDecorator.decorator.Base_Deco)
    
    Decorated a Class decorator (pass in method)

    method:
        __init__:
            params:
                pck: is pack (False)
                method_list: append method list [must be a list] ([])
                --other-- -> **params: append attribute dict(global)
        
        __call__:
            params:
                if self.pck, --other-- -> **c: append attribute dict(this)
                else cls: Decorated class
            
            return:
                function decofunc(cls):
                    params:
                        cls: Decorated class
                    
                    return:
                        cls: New class
                
                (if not self.pck, __call__ = decofunc)
        
    self params:
        method: append method list
        param: append attribute dict
        pck: is packed
    
    __str__ -> 'Decorator object: {self.__class__}'
    """
    def __init__(self, pck=False, method_list=[], **params):
        self._pck = pck
        self._param = params
        self._param_it = self._param
        self._method = method_list
        self._method_it = self._method


class Deco_iClass(Deco_aClass):
    """
    class NiceDecorator.decorator.Deco_iClass(NiceDecorator.decorator.Deco_aClass)

    Decorated a Class decorator (inheritance)

    method:
        __init__:
            --no params and return--

        __call__:
            params:
                cls: Decorated class

            return:
                cls: New class

        (definition static-add_attr:
            return:
                funcs: append method list [must be a list]
                attrs: append attribute dict [must be a dict]
        )

    self params:
        method: append method list
        param: append attribute dict
        pck: is packed

    __str__ -> 'Decorator object: {self.__class__}'
    """
    def __init__(self):
        self._method, self._param = self.add_attr()
        self._pck = False
        self._param_it = self._param
        self._method_it = self._method

    @staticmethod
    def add_attr():
        """
        static add_attr() (in NiceDecorator.decorator.Deco_iClass)

        empty

        return:
            funcs: []
            attrs: {}
        """
        attrs = {}
        funcs = []
        return funcs, attrs
# }Deco_for_class


# {
class Deco_Union(Base_Deco):
    """
    class NiceDecorator.decorator.Deco_Union(NiceDecorator.decorator.Base_Deco)

    Union the functions

    method:
        __init__:
            params:
                name: Union name

        __call__:
            params:
                --other-- -> *args, **kwargs: first func parameters

            return:
                if there are not any functions in self.run., None
                else res: last func return

        func:
            params:
                f: add function

            return:
                f: add function
        
        delete:
            params:
                f: remove function

            return:
                f: remove function

    self params:
        run: function list
        name: Union name
        union: union function

    __str__ -> 'Function union object: {self.name}'
    """
    def __init__(self, name=""):
        self.run: list[Callable] = []
        if name == "":
            self.name = self.__class__
        else:
            self.name = name

    def __call__(self, *args, **kwargs):
        return self._union(*args, **kwargs)

    def func(self, f):
        self.run.append(f)
        return f
    
    def delete(self, f):
        while f in self.run:
            self.run.remove(f)
        return f

    def _union(self, *args, **kwargs):
        if len(self.run) == 0:
            return None
        elif len(self.run) == 1:
            return self.run[0](*args, **kwargs)
        else:
            res = self.run[0](*args, **kwargs)
            for f in self.run[1:]:
                if isinstance(res, tuple):
                    if isinstance(res[0], tuple) and isinstance(res[1], dict): # (args, kwargs)
                        # Take the return value of the previous function as a parameter to the following function
                        res = f(*res[0], **res[1])
                    else:  # (args)
                        # Take the return value of the previous function as a parameter to the following function
                        res = f(*res)
                elif isinstance(res, dict):  # (kwargs)
                    # Take the return value of the previous function as a parameter to the following function
                    res = f(**res)
                else:
                    res = f(res)
            return res

    @property
    def union(self):
        """
        method union(self, *args, **kwargs) (in NiceDecorator.decorator.Deco_Union)

        Union function (Take the return value of the previous function as a parameter to the following function)

        params:
            ^self^
            --other-- -> *args, **kwargs: first function parameters

        return:
            res: last function return
        """
        return self._union

    @property
    def __str__(self):
        return f"Function union object: {self.name}"


class Deco_List(Base_Deco):
    """
    class NiceDecorator.decorator.Deco_List(NiceDecorator.decorator.Base_Deco)

    List the functions

    method:
        __init__:
            params:
                name: List name

        __call__:
            params:
                --other-- -> *args, **kwargs: unified parameters

            return:
                res: all func returns

        func:
            params:
                f: add function

            return:
                f: add function
        
        delete:
            params:
                f: remove function

            return:
                f: remove function

    self params:
        run: function list
        name: List name

    __str__ -> 'Function list object: {self.name}'
    """
    def __init__(self, name=""):
        self.run: list[Callable] = []
        if name == "":
            self.name = self.__class__
        else:
            self.name = name

    def __call__(self, *args, **kwargs):
        reses = []
        for c in self.run:
            reses.append(c(*args, **kwargs))

    def func(self, f):
        self.run.append(f)
        return f
    
    def delete(self, f):
        while f in self.run:
            self.run.remove(f)
        return f

    @property
    def __str__(self):
        return f"Function list object: {self.name}"
# }Deco_combination



"""
module NiceDecorator.set_param

A param getter and setter module

function:
    getparam(func): Getter param
    setparam(func, param_it, name): Setter param
"""
import inspect
import time
import re


def getparam(func):
    """
    function NiceDecorator.set_param.getparam(func)

    Getter param

    params:
        func: getter function

    return:
        _args: params
    """
    def have_name(x: inspect.Parameter):
        name = x.name
        if x.kind == inspect.Parameter.VAR_POSITIONAL:
            name = "*" + name
        elif x.kind == inspect.Parameter.VAR_KEYWORD:
            name = "**" + name
        return name

    _args = list(map(have_name, inspect.signature(func).parameters.values()))
    return _args


def setparam(func, param_it, name):
    """
    decorator NiceDecorator.set_param.setparam(func, param_it, name)

    Setter param

    params:
        func: decorated function
        param_it: params
        name: function class

    return:
        func: new function
        will_app: will add params
    """
    _args = getparam(func)
    argc = {}
    posarg_d = {}
    will_app = []

    if "func" in _args and name == "d":
        _args.remove("func")
        will_app.append("func")
    if "args" in _args and name == "b":
        _args.remove("args")
        will_app.append("args")
    if "kwargs" in _args and name == "b":
        _args.remove("kwargs")
        will_app.append("kwargs")
    if "res" in _args and name == "e":
        _args.remove("res")
        will_app.append("res")
    if "tram" in _args and name == "e":
        _args.remove("tram")
        will_app.append("tram")

    for _arg in _args:
        if _arg == "arg":
            for pmk, pmv in param_it.items():
                if re.match("^_[0-9]+__$", pmk):
                    posarg_d[int(pmk[1:-1])] = pmv
                elif re.match(f"^_{name}__[0-9]+_$", pmk):
                    posarg_d[int(pmk[3:-1])] = pmv
        else:
            for pmk, pmv in param_it.items():
                if pmk == _arg:
                    argc[pmk] = pmv
                    break
                elif pmk == f"_{name}_{_arg}_":
                    argc[pmk[3:-1]] = pmv
                    break

    def wrapper(**kw):
        assert set(kw.keys()) == set(will_app)
        posargs = list(posarg_d.values())
        res = func(*posargs, **argc, **kw)
        return res

    return wrapper, will_app



"""
package NiceDecorator

An easy and nice decorator package

modules:
    decorator: A decoration module(Decorator)
    objective: Package NiceDecorator 's check class module
    separator: Package NiceDecorator 's separator module
    timecount: Package NiceDecorator 's time count module
    set_param: A param getter and setter module
    model: A decorator model

function:
    PASS(): Pass function
    FUNCTION(): Tag decorated function (CANNOT CALL)

class:
    EMPTY(object): Empty class

help:
    PACKAGE_DIR: Package NiceDecorator 's dir
    CLASSES_MRO: Package NiceDecorator 's inheritance relationship of classes
    FUNCTION_MAGIC_PARAM_HELP: Package NiceDecorator 's magic paramters help
    WARNING: Package NiceDecorator 's used warning
"""


__all__ = ["decorator", "objective", "separator", "timecount", "model",
           "EMPTY",
           "FUNCTION", "PASS",
           "PACKAGE_DIR", "CLASSES_MRO", "FUNCTION_MAGIC_PARAM_HELP", "WARNING"]

PACKAGE_DIR = (
    """
NiceDecorator/
    __init__.py
        class EMPTY
        function PASS
        function FUNCTION
        help PACKAGE_PATH
        help CLASSES_MRO
        help FUNCTION_MAGIC_PARAM_HELP
        help WARNING
    decorator.py
        class Base_Deco
        group Deco_for_function:
            class Deco_Function
            class Deco_aFunction
            class Deco_iFunction
            class Deco_wFunction
        group Deco_for_class:
            class Deco_Class
            class Deco_aClass
            class Deco_iClass
        group Deco_combination
            class Deco_Union
            class Deco_List
    model.py
        group function:
            class Deco_Check
            class Deco_Expand
            class Deco_Timer
            class Deco_Timer_ns
        group class:
            class Deco_StrOpti
    objective.py
        function STRING
        function INTEGER
        function LIST
    separator.py
        function linesep
        function timesep
        function titlesep
    timecount.py
        group time_count
            function time_start
            function time_end
        group time_count_ns
            function time_start_ns
            function time_end_ns
    set_param.py
        function getparam
        function setparam
    """
)


CLASSES_MRO = (
    """
    decorator.Base_Deco
        decorator.Deco_Function
            decorator.Deco_aFunction
            decorator.Deco_iFunction
                model.Deco_Check
                model.Deco_Expand
                model.Deco_Timer
                model.Deco_Timer_ns
            decorator.Deco_wFunction
        decorator.Deco_Class
            decorator.Deco_aClass
            decorator.Deco_iClass
                decorator.Deco_StrOpti
    """
)


FUNCTION_MAGIC_PARAM_HELP = (
    """
Package NiceDecorator has 5 magic parameters, there are
'func', 'args', 'kwargs', 'res', 'tram'.

The 'func' parameter is used in the function definition section
and is automatically passed in to the decorated function. 
It can be modified. 
If used, MUST RETURN!
In 'model.Deco_Expand', 'func' is extended to the beginning and end of the function, 
but does not need to be returned. 
When used together with other magic parameters, 
it is placed before other magic parameters.

The 'args' parameter is used at the beginning of the function
and is automatically passed in as a positional parameter. 
It can also be modified. 
MUST RETURN!

The 'kwargs' parameter is used at the beginning of the function
and automatically passes in keyword parameters. It can also be modified. 
When passed in simultaneously with 'args' and/or' func '(from model.Deco_Expand), 
it should be placed after both. 
When returned simultaneously with 'args' and/or' tram ', 
placed after' args' and before 'tram'. 
If used, MUST RETURN!

The 'res' paramter is used at the end of the function
and automatically passes in the function return value. 
When used together with 'tram', 
it should be placed before 'tram'. 
If used, MUST RETURN!

The 'tram' parameter is used at the beginning
and end of the function to pass data. 
Returns 'tram' from the beginning of the function, 
located at the end. Received by the end of the function, 
should be at the end, and there is no need to return.
    """
)
WARNING = (
    """
Cannot add '/' to decorative functions
in each decoration function!

The parameters passed in when defining a decorator or decoration
can be received by each decoration function, 
but the positional parameters must be passed
in the format of '_'+passed in function properties ('d','b','e')+'__'+sort+'_'(specifying the passed in function)
or'_'+sort+'_'. 
Keyword parameters can be directly passed in, 
or the passed in function can be specified
in the format of '_'+passed in function properties ('d','b','e')+' __ '+parameter name+' _ '.

In this module, 
the function definition section is specified as'adef' (abbreviated as 'd'), 
the function start section is specified as 'begin' (abbreviated as 'b'), 
and the function end section is specified as 'end' (abbreviated as 'e').

In 'Deco_iClass', add methods and attributes uniformly in add_attr(), 
first return the list of add methods, 
and then return the add attribute dictionary.

In 'Deco_aFunction', 'Deco_iFunction',' Deco_aClass', 'Deco_iClass' and its subclasses, 
if pck is true, parameter passing is allowed, otherwise parameter passing is not allowed.

In 'Deco_aFunction', 'Deco_iFunction' and its subclasses 
are allowed to pass parameter keywords during decoration to indicate adding attributes 
(CANNOT have the same name as 'func' ! Positional parameters CANNOT be passed in!
Should not have the same name as magic parameters!)

In 'Deco_aClass', 'Deco_iClass' and its subclasses, if parameter passing is allowed, 
the positional parameter represents the adding method, 
and the keyword parameter represents the adding property (CANNOT be the same name as 'cls' !)
    """
)


def PASS():
    """
    function NiceDecorator.PASS()

    Pass function
    """
    pass


def FUNCTION():
    """
    function NiceDecorator.FUNCTION()

    Tag decorated function
    
    CANNOT CALL!
    """
    raise NotImplementedError("CANNOT CALL FUNCTION()!")


class EMPTY(object):
    """
    class NiceDecorator.EMPTY(object)

    Empty class
    """
    pass



"""
module NiceDecorator.timecount

Package NiceDecorator 's time count module

function:
    Group time_count:
        time_start(): Start time count
        time_end(res, tram): End time count and return
    Group time_count_ns:
        time_start_ns(): Start time count
        time_end_ns(res, tram): End time count and return
"""
import time


# {
def time_start():
    """
    function NiceDecorator.timecount.time_start() (Group time_count)

    Start time count

    return:
        $tram$: start time
    """
    tram = time.perf_counter()
    return tram


def time_end(res, tram):
    """
    function NiceDecorator.timecount.time_end() (Group time_count)

    End time count and return

    params:
        $res$: old response
        $tram$: start time

    return:
        $res$: new response
    """
    end = time.perf_counter()
    time_ = end - tram
    if res is None:
        res = time_
    else:
        res = (res, time_)
    return res
# } time_count

# {
def time_start_ns():
    """
    function NiceDecorator.timecount.time_start_ns() (Group time_count_ns)

    Start time count

    return:
        $tram$: start time
    """
    tram = time.perf_counter_ns()
    return tram


def time_end_ns(res, tram):
    """
    function NiceDecorator.timecount.time_end_ns() (Group time_count_ns)

    End time count and return

    params:
        $res$: old response
        $tram$: start time

    return:
        $res$: new response
    """
    end = time.perf_counter_ns()
    time_ = end - tram
    if res is None:
        res = time_
    else:
        res = (res, time_)
    return res
# } time_count_ns



"""
module NiceDecorator.model.

A decorator model

class:
    group function:
        Deco_Check(NiceDecorator.decorator.Deco_iFunction): A decorator Check class
        Deco_Expand(NiceDecorator.decorator.Deco_iFunction): Deco_Function 's Expand class
        Deco_Timer(NiceDecorator.decorator.Deco_iFunction): Deco_Function 's Timer Count Class
        Deco_Timer_ns(NiceDecorator.decorator.Deco_iFunction): Deco_Function 's Timer Count Class (Unit: nanoseconds)
    group class:
        Deco_StrOpti(NiceDecorator.decorator.Deco_iClass): A print Optimizer

"""
from sys import getsizeof
from typing import Iterable

from . import decorator, set_param, timecount


# {
class Deco_Check(decorator.Deco_iFunction):
    """
    class Deco_Check(NiceDecorator.decorator.NiceDecorator.Deco_iFunction)

    A decorator Check class

    method:
        __init__:
            params:
                pck: is pack (False)
                --other-- -> **kwargs: set some params(global)

        __call__:
            params:
                --other-- -> **c: set some params(part)

            return:
                function decofunc(func):
                    params:
                        func: Decorated function

                    return:
                        wrapper: Decorated function

                (if not pck, call = decofunc)

        (definition if_, true_, false_, begin, end)

    self params:
        wrap: (self.begin, NiceDecorator.FUNCTION, self.end)
        param: global params
        pck: is packed

    __str__ -> 'Decorator object: {self.__class__}'
    """
    def if_(self, str_con="True"):
        """
        method if_(self) (in NiceDecorator.model.Deco_Check)

        Return str_con bool

        params:
            str_con: condition

        return:
            bool_con: bool
        """
        return bool(eval(str_con))

    def true_(self):
        """
        method true_(self) (in NiceDecorator.model.Deco_Check)
        No args!

        run at True
        pass
        """
        pass

    def false_(self):
        """
        method false_(self) (in NiceDecorator.model.Deco_Check)
        No args!

        run at False
        pass
        """
        pass

    @property
    def adef(self):
        """
        method adef(self, func, con) (in NiceDecorator.model.Deco_Check)

        Check a condition

        if if_() return True, true_()
        else false_()
        CANNOT MODIFY!

        params:
            $func$: Decorated function
            con: condition

        return:
            $func$
        """

        def adef(func, con="True"):
            if self.if_(con):
                self.true_()
            else:
                self.false_()
            return func
        return adef


class Deco_Expand(decorator.Deco_iFunction):
    """
    class NiceDecorator.model.Deco_Expand(NiceDecorator.decorator.Deco_iFunction)

    Deco_Function 's Expand class

    method:
        __init__:
            params:
                pck: is pack (False)
                --other-- -> **kwargs: set some params(global)

        __call__:
            params:
                --other-- -> **c: set some params(part)

            return:
                function decofunc(func):
                    params:
                        func: Decorated function

                    return:
                        wrapper: Decorated function

                (if not pck, __call__ = decofunc)

        (definition adef_other,begin,end)

    self params:
        wrap: (self.begin, NiceDecorator.FUNCTION, self.end)
        param: global params
        pck: is packed

    __str__ -> 'Decorator object: {self.__class__}'
    """
    @property
    def adef(self):
        """
        method adef(self, func) (in NiceDecorator.model.Deco_Expand)

        Append 'func' to params

        CANNOT MODIFY!

        params:
            $func$: Decorated function

        return:
            $func$
        """
        def adef(func):
            then_do, wa = set_param.setparam(self._param_it, self.adef_other, "d")
            if "func" in wa:
                func = then_do(func=func)
            else:
                then_do()
            self._param_it["func"] = func
            return func
        return adef

    def adef_other(self):
        """
        method adef_other(self) (in NiceDecorator.model.Deco_Expand)

        pass
        """
        pass


class Deco_Timer(decorator.Deco_iFunction):
    """
    class NiceDecorator.model.Deco_Timer(NiceDecorator.decorator.Deco_iFunction)

    Deco_Function 's Timer Count Class

    method:
        __init__:
            params:
                pck: is pack (False)
                --other-- -> **kwargs: set some params(global)

        __call__:
            params:
                --other-- -> **c: set some params(part)

            return:
                function decofunc(func):
                    params:
                        func: Decorated function

                    return:
                        wrapper: Decorated function

                (if not pck, __call__ = decofunc)

        (definition adef)

    self params:
        wrap: (NiceDecorator.timecount.time_start, NiceDecorator.FUNCTION, NiceDecorator.timecount.time_end)
        param: global params
        pck: is packed

    __str__ -> 'Decorator object: {self.__class__}'

    """
    begin = staticmethod(timecount.time_start)
    end = staticmethod(timecount.time_end)


class Deco_Timer_ns(decorator.Deco_iFunction):
    """
    class NiceDecorator.model.Deco_Timer(NiceDecorator.decorator.Deco_iFunction)

    Deco_Function 's Timer Count Class (Unit: nanoseconds)

    method:
        __init__:
            params:
                pck: is pack (False)
                --other-- -> **kwargs: set some params(global)

        __call__:
            params:
                --other-- -> **c: set some params(part)

            return:
                function decofunc(func):
                    params:
                        func: Decorated function

                    return:
                        wrapper: Decorated function

                (if not pck, __call__ = decofunc)

        (definition adef)

    self params:
        wrap: (NiceDecorator.timecount.time_start_ns, NiceDecorator.FUNCTION, NiceDecorator.timecount.time_end_ns)
        param: global params
        pck: is packed

    __str__ -> 'Decorator object: {self.__class__}'

    """
    begin = staticmethod(timecount.time_start_ns)
    end = staticmethod(timecount.time_end_ns)
# } function


# {
class Deco_StrOpti(decorator.Deco_iClass):
    """
    class NiceDecorator.model.Deco_StrOpti(NiceDecorator.decorator.Deco_iClass)

    A print Optimizer

    method:
        __init__:
            --no params and return--

        __call__:
            params:
                cls: Decorated class

            return:
                cls: New class

        static-add_attr:
            CANNOT MODIFY!
            return:
                funcs: append method list [must be a list]
                attrs: append attribute dict [must be a dict]

    self params:
        method: append method list
        param: append attribute dict
        pck: is packed

    __str__ -> 'Decorator object: {self.__dict__}'
    """
    @staticmethod
    def add_attr():
        """
        static add_attr() (in NiceDecorator.model.Deco_StrOpti)

        Add a __str__ in cls

        return:
            funcs: [__str__:
                params:
                    ^self^
                return:
                    res_str: Processing Long Strings
                ]
            attrs: {}
        """
        def __str__(self):
            res_str = "\n"
            line = [f"{self.__class__} object: "]
            for name, value in self.__dict__.items():
                line.append(f"    {name}: {value} ")

            def size(x):
                sum_size = 0
                if isinstance(x, str):
                    a = getsizeof(x)
                    b = getsizeof(x * 5)
                    sum_size = (b-a) // 4
                elif isinstance(x, Iterable):
                    for item in x:
                        sum_size += size(item)
                else:
                    sum_size = getsizeof(x)
                return sum_size

            line.append(f"size: {size(self)}, id: {id(self)}")
            res_str.join(line)
            return res_str
        funcs = [__str__]
        attrs = {}
        return funcs, attrs
# } class



from setuptools import setup

with open(r"NiceDecorator\README.md", "r") as f:
    long_description = f.read()

setup(name='NiceDecorator',  # 包名
      version='1.1.1',  # 版本号
      description='An easy and nice decorator package',
      long_description=long_description,
      long_description_content_type="text/markdown",
      license='MIT License',
      packages=["NiceDecorator"],
      keywords='decorator',
      python_requires='>=3.10, <4',
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Programming Language :: Python :: 3.12',
      ],
    )



"""
module NiceDecorator.objective.

Package NiceDecorator 's check class module

functions:
    STRING(args, kwargs): A batch force string conversion function
    INTEGER(args, kwargs): A batch force integer conversion function
    LIST(args, kwargs): A batch force list conversion function
"""

from typing import List, Dict


def STRING(args: List, kwargs: Dict):
    """
    function NiceDecorator.objective.STRING(args, kwargs)

    A batch force string conversion function

    if param is list or tuple, return " ".join(param)
    if param is dict, return " ".join(param.values())
    if param is set, raise
    else return str(param)

    params:
        $args$: the function 's nameless params [must be a list]
        $kwargs$: the function 's with name params [must be a dict]

    return:
        $args$: str nameless params [list]
        $kwargs$: str with name params [dict]
    """
    new_arg = []
    new_kwarg = {}

    def _str_(x):
        if isinstance(x, tuple):
            _1 = x[0]
            _2 = x[1]
            if isinstance(_2, list) or isinstance(_2, tuple):
                new_kwarg[_1] = " ".join(_2)
            elif isinstance(_2, dict):
                new_kwarg[_1] = " ".join(_2.values())
            elif isinstance(_2, set):
                raise TypeError("cannot get 'set' to 'str'")
            else:
                new_kwarg[_1] = str(_2)
        else:
            _ = x
            if isinstance(_, list) or isinstance(_, tuple):
                new_arg.append(" ".join(_))
            elif isinstance(_, dict):
                new_arg.append(" ".join(_.values()))
            elif isinstance(_, set):
                raise TypeError("cannot get 'set' to 'str'")
            else:
                new_arg.append(str(_))

    map(_str_, args, kwargs.items())
    args, kwargs = new_arg, new_kwarg
    return args, kwargs


def INTEGER(args: List, kwargs: Dict):
    """
    function NiceDecorator.objective.INTEGER(args, kwargs)

    A batch force integer conversion function

    if param is list or tuple or set, return length of param
    if param is dict, return number of key value pairs
    if param is str:
        if can int(param), return int(param)
        else return len(param)
    else return int(param)

    params:
        $args$: the function 's nameless params [must be a list]
        $kwargs$: the function 's with name params [must be a dict]

    return:
        $args$: int nameless params [list]
        $kwargs$: int with name params [dict]
    """
    new_arg = []
    new_kwarg = {}
    
    def Len_Dict(d: Dict):
        _len_ = 0
        for _ in d.values():
            if isinstance(_, dict):
                _len_ += Len_Dict(_)
            else:
                _len_ += 1
        return _len_

    def _int_(x):
        if isinstance(x, tuple):
            _1 = x[0]
            _2 = x[1]
            if isinstance(_2, list) or isinstance(_2, tuple) or isinstance(_2, set):
                new_kwarg[_1] = len(_2)
            elif isinstance(_2, dict):
                new_kwarg[_1] = Len_Dict(_2)
            elif isinstance(_2, str):
                try:
                    new_kwarg[_1] = int(_2)
                except ValueError:
                    new_kwarg[_1] = len(_2)
            else:
                new_kwarg[_1] = int(_2)
        else:
            _ = x
            if isinstance(_, list) or isinstance(_, tuple) or isinstance(_, set):
                new_arg.append(len(_))
            elif isinstance(_, dict):
                new_arg.append(Len_Dict(_))
            elif isinstance(_, str):
                try:
                    new_arg.append(int(_))
                except ValueError:
                    new_arg.append(len(_))
            else:
                new_arg.append(int(_))

    map(_int_, args, kwargs.items())
    args, kwargs = new_arg, new_kwarg
    return args, kwargs


def LIST(args: List, kwargs: Dict):
    """
    function NiceDecorator.objective.LIST(args, kwargs)

    A batch force list conversion function

    if param is str, return it split(" ")
    if param is int, return list(length = param)
    if param is dict, return it values()
    else return list(param)

    params:
        $args$: the function 's nameless params [must be a list]
        $kwargs$: the function 's with name params [must be a dict]

    return:
        $args$: list nameless params [list]
        $kwargs$: list with name params [dict]
    """
    new_arg = []
    new_kwarg = []

    def _list_(x):
        if isinstance(x, tuple):
            _1 = x[0]
            _2 = x[1]
            if isinstance(_2, str):
                new_kwarg[_1] = _2.split(" ")
            elif isinstance(_2, int):
                new_kwarg[_1] = list(range(_2))
            elif isinstance(_2, dict):
                new_kwarg[_1] = list(_2.values())
            else:
                new_kwarg[_1] = list(_2)
        else:
            _ = x
            if isinstance(_, str):
                new_arg.append(_.split(" "))
            elif isinstance(_, int):
                new_arg.append(list(range(_)))
            elif isinstance(_, dict):
                new_arg.append(list(_.values()))
            else:
                new_arg.append(list(_))

    map(_list_, args, kwargs.items())
    args, kwargs = new_arg, new_kwarg
    return args, kwargs



"""
module NiceDecorator.separator

Package NiceDecorator 's separator module

function:
    linesep(_star, _len): print a line separator
    timesep(_method): print time for separator
    titlesep(_title, _star, _len): print a line separator with title
"""
import datetime
import time


def linesep(_star: str = "*", _len: int = 10):
    """
    function NiceDecorator.separator.linesep(_star, _len)

    print a line separator

    params:
        _star: separator string [must be a str] ("*")
        _len: separator len [must be a int] (10)
    """
    print(_star*_len)


def timesep(_method: str = "DATETIME"):
    """
    function NiceDecorator.separator.timesep(_rou, _method)

    print time for separator

    params:
        _method: time show method [must be str] ("DATETIME")
    """
    if _method == "DATETIME":
        time_ = datetime.datetime.now().strftime("%Y.%m.%d %H:%M:%S.%f")
    elif _method == "CPU":
        time_ = time.time()
    else:
        raise ValueError(_method)

    print(time_)


def titlesep(_title: str, _star: str = "-", _len: int = 5):
    """
    function NiceDecorator.separator.titlesep(_title, _star, _len)

    print a line separator with title

    params:
        _title: title string [must be a str]
        _star: separator string [must be a str] ("-")
        _len: separator len [must be a int] (5)
    """
    print(f"{_star*_len}{_title}{_star*_len}")
